#include "sdl.h"
#include <vector>
#include <cstdio>
#include <cstdlib>

// Factor de conversion grados a radianes:
static constexpr float pi= M_PI;
static constexpr float factor= pi/180.0;
// Gráfico de nave, 5 puntos, 4 segmentos:
static const std::vector<sdl::FPoint> nave = { {15.0, 0.0}, {-15.0, -9.0}, {-11.0, 0.0}, {-15.0, 9.0}, {15.0, 0.0} };
static const std::vector<sdl::FPoint> estela = { {-16.0, -4.0}, {-25.0, 0.0}, {-16.0, 4.0} };
static const std::vector<std::vector<sdl::FPoint>> asteroide = {
    {{-5.0, -18.0}, {10.0, -16.0}, {19.0, -3.0}, {10.0, 7.0}, {11.0, 15.0}, {-5.0, 19.0}, {-18.0,8.0}, {-14.0, -7.0}, {-5.0, -18.0} },
    {{-10.0, -14.0}, {2.0, -14.0}, {11.0, 1.0}, {14.0, 1.0}, {8.0, 9.0}, {-2.0, 14.0}, {-9.0,6.0}, {-14.0, 4.0}, {-9.0, -5.0}, {-10.0,-14.0} },
    {{-4.0, -10.0}, {2.0, -10.0}, {9.0, -6.0}, {5.0, 1.0}, {8.0, 5.0}, {0.0, 9.0}, {-5.0,-2.0}, {-7.0, -4.0}, {-4.0, -10.0} }};
static const std::vector<std::vector<sdl::Shape>> forma = {
    { sdl::FCircle(0, 0, 19) },
    { sdl::FCircle(0, 0, 14) },
    { sdl::FCircle(0, 0, 10) }
};

enum tipoAsteroide {
    grande=0,
    mediano,
    pequeno,
    nada
};

class Nave {
private:
    std::vector<sdl::FPoint> nave2;
    std::vector<sdl::FPoint> estela2;
    sdl::FPoint posicion;
    sdl::FPoint velocidad;
    void trasladar(float alfa, sdl::FPoint &desplazamiento, sdl::FPoint &escala) {
        for(size_t i = 0; i < nave.size(); i++) {
            nave2[i] = nave[i].Move(alfa, desplazamiento, escala);
        }
        for(size_t i = 0; i < estela.size(); i++) {
            estela2[i] = estela[i].Move(alfa, desplazamiento, escala);
        }
    }

public:
    Nave(const sdl::FPoint &pos, const float alfa) : posicion(pos), velocidad(0.0, 0.0) {
        nave2.resize(nave.size());
        estela2.resize(estela.size());
    }
    void Actualizar(Uint64 tick, float acelerar, float alfa, sdl::FPoint escala) {
        velocidad.X() += acelerar*cos(factor*alfa);
        velocidad.Y() += -acelerar*sin(factor*alfa);
        posicion.X() += (velocidad.X()*(tick)/1000.0);
        posicion.Y() += (velocidad.Y()*(tick)/1000.0);
        if(posicion.X()>635) posicion.X() -=640;
        if(posicion.X()<5) posicion.X() +=640;
        if(posicion.Y()>475) posicion.Y() -=480;
        if(posicion.Y()<5) posicion.Y() +=480;
        // Trasladar nave:
        trasladar(factor*alfa, posicion, escala);
    }
    std::vector<sdl::FPoint> &PuntosNave() { return nave2; }
    std::vector<sdl::FPoint> &PuntosEstela() { return estela2; }
    sdl::FPoint Posicion() { return posicion; }
};

class Asteroide {
private:
    std::vector<sdl::FPoint> asteroide2;
    tipoAsteroide tipo;
    sdl::FPoint posicion;
    sdl::FPoint velocidad;
    float alfa; // En radianes
    float vRotacion; // En radianes/tick
    sdl::Hitbox hitbox;
    //bool borrar;
    void Trasladar(float alfa, sdl::FPoint &desplazamiento, sdl::FPoint &escala) {
    // x' = x * x" * cos(α) + y * y" * sin(α) + dx
    // y' = -x * x" * sin(α) + y * y" * cos(α) + dy
        for(size_t i = 0; i < asteroide2.size(); i++) {
            asteroide2[i] = asteroide[tipo][i].Move(alfa, desplazamiento, escala);
        }
}
public:
    Asteroide(const sdl::FPoint &pos, tipoAsteroide t) : tipo(t), posicion(pos), alfa(360.0*std::rand()/RAND_MAX) {
        asteroide2.resize(asteroide[tipo].size());
        float v = 100.0*std::rand()/RAND_MAX;
        alfa = 2.0*pi*std::rand()/RAND_MAX;
        velocidad = {v*(float)cos(alfa), v*(float)sin(alfa)};
        vRotacion = -5.0+10.0*std::rand()/RAND_MAX;
        for(size_t i=0; i<forma[tipo].size(); i++) hitbox.InsertShape(forma[tipo][i]);
        hitbox.ResizeT();
    }
    void Actualizar(Uint64 tick, sdl::FPoint escala) {
        posicion.X() += (velocidad.X()*(tick)/1000.0);
        posicion.Y() += (velocidad.Y()*(tick)/1000.0);
        alfa += vRotacion *(tick)/1000.0;
        if(alfa > 2*pi) alfa -= 2*pi;
        if(alfa < 0) alfa += 2*pi;
        if(posicion.X()>645) posicion.X() -=650;
        if(posicion.X()<-5) posicion.X() +=650;
        if(posicion.Y()>485) posicion.Y() -=490;
        if(posicion.Y()<-5) posicion.Y() +=490;
        // Trasladar asteroide:
        Trasladar(alfa, posicion, escala);
        // Trasladar hitbox:
        hitbox.Move(alfa, posicion, escala);
    }
    tipoAsteroide Tipo() { return tipo; }
    sdl::FPoint &Posicion() { return posicion; }
    sdl::FPoint &Velocidad() { return velocidad; }
    bool Colision(Asteroide &a) {
        return hitbox.Collision(a.hitbox);
    }
    std::vector<sdl::FPoint> &Puntos() { return asteroide2; }
};

class Asteroides : public sdl::Game {
private:
    sdl::FPoint escala; // Escala del juego (ajustada a una resolución de 640x480
    std::vector<sdl::FPoint> estrellas;
    Nave nave;
    std::vector<Asteroide> asteroides;
    float acelerar;
    float alfa;
    void RebotarAsteroides(int a, int b);

public:
    Asteroides(sdl::Window& window, sdl::Renderer& renderer) : sdl::Game(window, renderer), nave({320.0, 240.0}, 0.0),
        acelerar(0.0), alfa(0.0) {
        static int w, h;
        window.GetSize(&w, &h);
        escala = sdl::FPoint(w/640.0, h/480.0);
        for(int i=0; i<500; i++) {
            estrellas.push_back({static_cast<float>(w*rand()/RAND_MAX), static_cast<float>(h*rand()/RAND_MAX)});
        }
    }
    const sdl::FPoint &Escala() { return escala; }
    void Init();
    void Events();
    void Update();
    void Render();
};

void Asteroides::Events() {
    static bool izq;
    static bool der;
    static bool arr;

    while(event.Poll()) {
        switch(event.Type()) {
        case SDL_KEYDOWN:
            if(event.Key().keysym.sym == SDLK_LEFT)  { izq = true; }
            if(event.Key().keysym.sym == SDLK_RIGHT) { der = true; }
            if(event.Key().keysym.sym == SDLK_UP)    { arr = true; }
            if(event.Key().keysym.sym == SDLK_ESCAPE) {
                quit = true; }
            break;
        case SDL_KEYUP:
            if(event.Key().keysym.sym == SDLK_LEFT)  { izq = false; }
            if(event.Key().keysym.sym == SDLK_RIGHT) { der = false; }
            if(event.Key().keysym.sym == SDLK_UP)    { arr = false; }
            break;
        }
    }
    acelerar = 0.0;
    if(izq && !der) alfa -= tick/10.0;
    if(!izq && der) alfa += tick/10.0;
    if(alfa < 0) alfa+=360.0;
    if(alfa >= 360.0) alfa-=360.0;
    if(arr) acelerar = tick/15.0;
}

void Asteroides::Render() {
    // Renderizado:
    renderer.SetDrawColor(0, 0, 0, 255);
    renderer.Clear();
    renderer.SetDrawColor(255, 255, 255, 255);
    for(size_t i=0; i<estrellas.size(); i++)
        renderer.DrawPoint(estrellas[i]);
    renderer.DrawLinesF(nave.PuntosNave());
    if(acelerar != 0.0) {
        renderer.SetDrawColor(255, 255, 0, 255);
        renderer.DrawLinesF(nave.PuntosEstela());
    }
    // Asteroides:
    for(size_t i=0; i<asteroides.size(); i++) {
        renderer.SetDrawColor(0,255,0,255);
        renderer.DrawLinesF(asteroides[i].Puntos());
    }
    renderer.Present();
}

void Asteroides::Update() {
    nave.Actualizar(tick, acelerar, alfa, escala);
    for(size_t i=0; i<asteroides.size(); i++) asteroides[i].Actualizar(tick, escala);
    // Detectar colisiones
    for(size_t i=0; i<asteroides.size()-1; i++) {
        for(size_t j=i+1; j<asteroides.size(); j++) {
            if(asteroides[i].Colision(asteroides[j])) {
                RebotarAsteroides(i, j);
            }
        }
    }
}

void Asteroides::RebotarAsteroides(int a, int b) {
    sdl::FPoint V1, V2;
    float theta;
    float m1,m2;

    m1 = (asteroides[a].Tipo()==grande) ? 4.0: (asteroides[a].Tipo()==mediano) ? 2.0:1.0;
    m2 = (asteroides[b].Tipo()==grande) ? 4.0: (asteroides[b].Tipo()==mediano) ? 2.0:1.0;
    // Calcular el ángulo normal:
    // θ = atan2(y2-y1, x2-x1)
    theta = SDL_atan2f(asteroides[b].Posicion().Y()-asteroides[a].Posicion().Y(),
                       asteroides[b].Posicion().X()-asteroides[a].Posicion().X());
    // Calcular componentes normal y transversal de las velocidades
    // Vn =  Vx · cos(θ) + Vy · sin(θ)
    // Vt = -Vx · sin(θ) + Vy · cos(θ)
    V1.X() =  asteroides[a].Velocidad().X()*cos(theta)+asteroides[a].Velocidad().Y()*sin(theta);
    V1.Y() = -asteroides[a].Velocidad().X()*sin(theta)+asteroides[a].Velocidad().Y()*cos(theta);
    V2.X() =  asteroides[b].Velocidad().X()*cos(theta)+asteroides[b].Velocidad().Y()*sin(theta);
    V2.Y() = -asteroides[b].Velocidad().X()*sin(theta)+asteroides[b].Velocidad().Y()*cos(theta);
    // Calcular nuevas direcciones
    // v1fn = (2·v2in·m2+v1in·(m1-m2))/(m1+m2)
    // v2fn = (2·v1in·m1+v2in·(m2-m1))/(m1+m2)
    // v1ft = v1it
    // v2ft = v2it
    float v2n = (2*V1.X()*m1+V2.X()*(m2-m1))/(m1+m2);
    float v1n = (2*V2.X()*m2+V1.X()*(m1-m2))/(m1+m2);
    V1.X() = v1n;
    V2.X() = v2n;
    // Recalcular componentes ortogonales:
    // Vx = Vn · cos(θ) - Vt · sin(θ)
    // Vy = Vn · sin(θ) + Vt · cos(θ)
    asteroides[a].Velocidad().X() = V1.X()*cos(theta)-V1.Y()*sin(theta);
    asteroides[a].Velocidad().Y() = V1.X()*sin(theta)+V1.Y()*cos(theta);
    asteroides[b].Velocidad().X() = V2.X()*cos(theta)-V2.Y()*sin(theta);
    asteroides[b].Velocidad().Y() = V2.X()*sin(theta)+V2.Y()*cos(theta);
}

void Asteroides::Init() {
    // Añadir 5 asteroides grandes
    for(int i=0; i<5; i++) {
        asteroides.push_back(Asteroide({static_cast<float>(640.0*rand()/RAND_MAX), static_cast<float>(480.0*rand()/RAND_MAX)}, grande));
    }
}

int main( int argc, char * argv[] ) {
    sdl::SDL2 sdl(SDL_INIT_EVERYTHING);
    sdl::Window window ("Asteroides", 800, 600, (SDL_WindowFlags)(SDL_WINDOW_SHOWN | SDL_WINDOW_BORDERLESS));
    sdl::Renderer renderer(window, SDL_RENDERER_ACCELERATED);

    Asteroides asteroides(window, renderer);

    asteroides.Run();
    return 0;
}
